package com.samelement.flutter_in_app_update;

import android.app.Activity;
import android.app.Application;
import android.content.Context;
import android.content.Intent;
import android.content.IntentSender;
import android.os.Bundle;
import android.util.Log;

import androidx.annotation.NonNull;
import androidx.annotation.Nullable;

import com.google.android.play.core.appupdate.AppUpdateInfo;
import com.google.android.play.core.appupdate.AppUpdateManager;
import com.google.android.play.core.appupdate.AppUpdateManagerFactory;
import com.google.android.play.core.install.InstallStateUpdatedListener;
import com.google.android.play.core.install.model.AppUpdateType;
import com.google.android.play.core.install.model.InstallStatus;
import com.google.android.play.core.install.model.UpdateAvailability;
import com.google.android.play.core.tasks.Task;

import java.util.HashMap;
import java.util.Map;

import io.flutter.embedding.engine.plugins.FlutterPlugin;
import io.flutter.embedding.engine.plugins.activity.ActivityAware;
import io.flutter.embedding.engine.plugins.activity.ActivityPluginBinding;
import io.flutter.plugin.common.MethodCall;
import io.flutter.plugin.common.MethodChannel;
import io.flutter.plugin.common.MethodChannel.MethodCallHandler;
import io.flutter.plugin.common.MethodChannel.Result;
import io.flutter.plugin.common.PluginRegistry;
import io.flutter.plugin.common.PluginRegistry.Registrar;

/**
 * FlutterInAppUpdatePlugin
 */
public class FlutterInAppUpdatePlugin implements FlutterPlugin, MethodCallHandler,
        PluginRegistry.ActivityResultListener, ActivityAware, Application.ActivityLifecycleCallbacks {
    /// The MethodChannel that will the communication between Flutter and native Android
    ///
    /// This local reference serves to register the plugin with the Flutter Engine and unregister it
    /// when the Flutter Engine is detached from the Activity
    private MethodChannel channel;

    private static Context context;
    private Activity activity;
    private AppUpdateManager appUpdateManager;
    private InstallStateUpdatedListener listener;
    private Result updateResult;
    private int updateType;
    private boolean isUpdateAvailable = false;

    private final int HIGH_PRIORITY_UPDATE = 5;
    private final int LOW_PRIORITY_UPDATE = 1;
    private final int MY_REQUEST_CODE = 11998733;

    @Override
    public void onAttachedToEngine(@NonNull FlutterPluginBinding flutterPluginBinding) {
        channel = new MethodChannel(flutterPluginBinding.getBinaryMessenger(), "flutter_in_app_update");
        channel.setMethodCallHandler(this);

        context = flutterPluginBinding.getApplicationContext();
    }

    // This static function is optional and equivalent to onAttachedToEngine. It supports the old
    // pre-Flutter-1.12 Android projects. You are encouraged to continue supporting
    // plugin registration via this function while apps migrate to use the new Android APIs
    // post-flutter-1.12 via https://flutter.dev/go/android-project-migration.
    //
    // It is encouraged to share logic between onAttachedToEngine and registerWith to keep
    // them functionally equivalent. Only one of onAttachedToEngine or registerWith will be called
    // depending on the user's project. onAttachedToEngine or registerWith must both be defined
    // in the same class.
    public static void registerWith(Registrar registrar) {
        final MethodChannel channel = new MethodChannel(registrar.messenger(), "flutter_in_app_update");
        channel.setMethodCallHandler(new FlutterInAppUpdatePlugin());

        context = registrar.context();
    }

    @Override
    public void onMethodCall(@NonNull MethodCall call, @NonNull Result result) {
        if (call.method.equals("checkForUpdate")) {
            int stalenessDays = call.argument("staleness_days");
            updateResult = result;

            // Creates instance of the manager.
            appUpdateManager = AppUpdateManagerFactory.create(context);

            // Returns an intent object that you use to check for an update.
            Task<AppUpdateInfo> appUpdateInfoTask = appUpdateManager.getAppUpdateInfo();

            // Checks that the platform will allow the specified type of update.
            appUpdateInfoTask.addOnSuccessListener(appUpdateInfo -> {
                if (appUpdateInfo.updateAvailability() == UpdateAvailability.UPDATE_AVAILABLE) {
                    isUpdateAvailable = true;

                    // Check staleness
                    if ((appUpdateInfo.clientVersionStalenessDays() != null &&
                            appUpdateInfo.clientVersionStalenessDays() >= stalenessDays) ||
                            determinePriority(appUpdateInfo.availableVersionCode()) == HIGH_PRIORITY_UPDATE) {
                        // Will be immediate update
                        if (appUpdateInfo.isUpdateTypeAllowed(AppUpdateType.IMMEDIATE)) {
                            updateType = AppUpdateType.IMMEDIATE;
                            performImmediateUpdate(appUpdateInfo);
                        }
                    } else {
                        // Will be flexible update
                        if (appUpdateInfo.isUpdateTypeAllowed(AppUpdateType.FLEXIBLE)) {
                            updateType = AppUpdateType.FLEXIBLE;
                            performFlexibleUpdate(appUpdateInfo);
                        }
                    }

                }

                returnResult(false);
            }).addOnFailureListener(e -> {
                Log.e("FlutterInAppUpdate", e.getMessage());
                isUpdateAvailable = false;
                updateType = AppUpdateType.IMMEDIATE;
                returnResult(false);
            });

        } else {
            result.notImplemented();
        }
    }

    @Override
    public void onDetachedFromEngine(@NonNull FlutterPluginBinding binding) {
        channel.setMethodCallHandler(null);
    }

    @Override
    public boolean onActivityResult(int requestCode, int resultCode, Intent data) {
        if (requestCode == MY_REQUEST_CODE) {
            if (resultCode != Activity.RESULT_OK) {
                returnResult(false);
            } else {
                returnResult(true);
            }
        }

        return false;
    }

    private void returnResult(boolean success) {
        Map<String, Object> map = new HashMap<String, Object>();
        map.put("isUpdateSuccess", success);
        map.put("updateType", updateType);
        map.put("isUpdateAvailable", isUpdateAvailable);

        if (updateResult != null) updateResult.success(map);
    }

    // 1 .... 5
    private int determinePriority(int updateVersionCode) {
        if (updateVersionCode % 2 == 1) return HIGH_PRIORITY_UPDATE;

        return LOW_PRIORITY_UPDATE;
    }

    private void performImmediateUpdate(AppUpdateInfo appUpdateInfo) {
        try {
            appUpdateManager.startUpdateFlowForResult(
                    appUpdateInfo,
                    AppUpdateType.IMMEDIATE,
                    activity,
                    MY_REQUEST_CODE);
        } catch (IntentSender.SendIntentException e) {
            e.printStackTrace();
        }
    }

    private void performFlexibleUpdate(AppUpdateInfo appUpdateInfo) {
        try {
            // Create a listener to track request state updates.
            listener = state -> {
                // (Optional) Provide a download progress bar.
                if (state.installStatus() == InstallStatus.DOWNLOADED) {
                    // After the update is downloaded, show a notification
                    // and request user confirmation to restart the app.
                    appUpdateManager.completeUpdate();
                }

                if (state.installStatus() == InstallStatus.INSTALLED) {
                    appUpdateManager.unregisterListener(listener);
                }

                if (state.installStatus() == InstallStatus.CANCELED || state.installStatus() == InstallStatus.FAILED) {
                    returnResult(false);
                }
                // Log state or install the update.
            };

// Before starting an update, register a listener for updates.
            appUpdateManager.registerListener(listener);

            appUpdateManager.startUpdateFlowForResult(
                    // Pass the intent that is returned by 'getAppUpdateInfo()'.
                    appUpdateInfo,
                    // Or 'AppUpdateType.FLEXIBLE' for flexible updates.
                    AppUpdateType.FLEXIBLE,
                    // The current activity making the update request.
                    activity,
                    // Include a request code to later monitor this update request.
                    MY_REQUEST_CODE);
        } catch (IntentSender.SendIntentException e) {
            e.printStackTrace();
        }
    }

    @Override
    public void onAttachedToActivity(@NonNull ActivityPluginBinding binding) {
        activity = binding.getActivity();
    }

    @Override
    public void onDetachedFromActivityForConfigChanges() {

    }

    @Override
    public void onReattachedToActivityForConfigChanges(@NonNull ActivityPluginBinding binding) {

    }

    @Override
    public void onDetachedFromActivity() {

    }

    @Override
    public void onActivityCreated(@NonNull Activity activity, @Nullable Bundle savedInstanceState) {

    }

    @Override
    public void onActivityStarted(@NonNull Activity activity) {

    }

    @Override
    public void onActivityResumed(@NonNull Activity activity) {
        appUpdateManager
                .getAppUpdateInfo()
                .addOnSuccessListener(appUpdateInfo -> {
                    // If the update is downloaded but not installed,
                    // notify the user to complete the update.
                    // flexible
                    if (appUpdateInfo.installStatus() == InstallStatus.DOWNLOADED) {
                        appUpdateManager.completeUpdate();
                    }

                    if (appUpdateInfo.installStatus() == InstallStatus.INSTALLED) {
                        appUpdateManager.unregisterListener(listener);
                    }

                    // immediate
                    if (appUpdateInfo.updateAvailability()
                            == UpdateAvailability.DEVELOPER_TRIGGERED_UPDATE_IN_PROGRESS) {
                        // If an in-app update is already running, resume the update.
                        performImmediateUpdate(appUpdateInfo);
                    }
                });
    }

    @Override
    public void onActivityPaused(@NonNull Activity activity) {

    }

    @Override
    public void onActivityStopped(@NonNull Activity activity) {

    }

    @Override
    public void onActivitySaveInstanceState(@NonNull Activity activity, @NonNull Bundle outState) {

    }

    @Override
    public void onActivityDestroyed(@NonNull Activity activity) {
        appUpdateManager.unregisterListener(listener);
    }
}
